//
// Lighthouse3D.com
//
//
// This demo was built for learning purposes only.
// Some code could be severely optimised, but I tried to
// keep as simple and clear as possible.
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//
// If you do use it I would love to hear about it. Just post a comment
// at Lighthouse3D.com

// Have Fun :-)

//
// Program for lecturing purposes by Pavel Drobek (paveldrobek@mail.muni.cz)
// Slight modifications by Marek Vinkler (xvinkl@fi.muni.cz).
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//


#include <cmath>

// Include GLEW to access OpenGL 3.3 functions
#include <GL/glew.h>

// GLUT is the toolkit to interface with the OS
#include <GL/freeglut.h>

#include "rayTracer.h"

// Use Very Simple Libs
#include "vsMathLib.h"
#include "vsShaderLib.h"
#include "vsLogLib.h"

//----------------------------------------------------------------
// TODO:
// 1) Inspect functions "setupShaders", "init" and "renderScene".
// 2) Follow the comments in shaders to implement the Whitted style ray tracing.
// 3) Create a scene of your own in function "init".
//----------------------------------------------------------------

VSMathLib *vsml;
VSShaderLib shader, initShader, finalShader;

VSLogLib mLogError, mLogInfo;

// Vao
GLuint vao;

// Width and height of generated image
int width = 512;
int height = 512;

// Texture Data for planes, spheres and materials
GLuint dataID;
GLuint materialsID;

// Texture Data for framebuffer
GLuint raysOriginsTexture[2], raysDirectionsTexture[2], colorTexture[2];

// Our data framebuffers
GLuint dataFramebuffer[2];

int reflections = 2;
float depth = 1.2f;
float reflectionCoef = 2.42f;

// ------------------------------------------------------------
//
// Reshape Callback Function
//

void changeSize(int w, int h)
{
	// Prevent a divide by zero, when window is too short
	if(h == 0)
		h = 1;

	// Set the viewport to be the entire window
	glViewport(0, 0, w, h);

	// Save information about width and height
	width = w;
	height = h;

	vsml->loadIdentity(VSMathLib::PROJECTION);

	vsml->ortho(-1.f, 1.f, -1.f, 1.f);
}


// ------------------------------------------------------------
//
// Render stuff
//

void bindFramebufferTargets(GLuint framebuffer)
{
	// Bind created buffer
	glBindFramebuffer(GL_FRAMEBUFFER, framebuffer);

	// Attach connected textures with draw buffers
	GLenum DrawBuffers[3] = {GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1, GL_COLOR_ATTACHMENT2};
	glDrawBuffers(3, DrawBuffers);
}

void renderScene(void)
{
	bindFramebufferTargets(dataFramebuffer[0]);

	glClear(GL_COLOR_BUFFER_BIT);

	vsml->loadIdentity(VSMathLib::VIEW);
	vsml->loadIdentity(VSMathLib::MODEL);

	// Set camera looking at our rendered triangles
	// Any z-position greater than 0 is valid due to orthogonal projection
	vsml->lookAt(0.f,0.f,1.f, 0,0,0, 0,1,0);

	// Use init shader
	glUseProgram(initShader.getProgramIndex());
	
	// Send matrices to uniform buffer
	vsml->matricesToGL();

	// Send width and height information to initShader
	initShader.setUniform("width", (float)width);
	initShader.setUniform("height", (float)height);
	initShader.setUniform("depth", depth);

	// Render VAO for initShader
	glBindVertexArray(vao);
	glDrawElements(GL_TRIANGLES, faceCount*3, GL_UNSIGNED_INT, 0);

	for(int i = 0; i < reflections; i++)
	{
		bindFramebufferTargets(dataFramebuffer[1 - (i % 2)]);

		// Bind and activate textures on positions 0 and 1
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, raysOriginsTexture[i % 2]);
		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, raysDirectionsTexture[i % 2]);
		glActiveTexture(GL_TEXTURE2);
		glBindTexture(GL_TEXTURE_2D, colorTexture[i % 2]);

		// Use our ray tracer shader
		glUseProgram(shader.getProgramIndex());
	
		// Render VAO for rayTracer shader
		glBindVertexArray(vao);
		glDrawElements(GL_TRIANGLES, faceCount*3, GL_UNSIGNED_INT, 0);
	}

	// Unbind our data framebuffer and render result on screen
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	glActiveTexture(GL_TEXTURE2);
	glBindTexture(GL_TEXTURE_2D, colorTexture[reflections % 2]);
	glUseProgram(finalShader.getProgramIndex());
	

	// Render VAO
	glBindVertexArray(vao);
	glDrawElements(GL_TRIANGLES, faceCount*3, GL_UNSIGNED_INT, 0);

	// Swap buffers
	glutSwapBuffers();
}


// ------------------------------------------------------------
//
// Events from the Keyboard
//

void processKeys(unsigned char key, int xx, int yy)
{
	switch(key)
	{
		case 27:
			glutLeaveMainLoop();
			break;
		case 'l':
			reflections += 1;
			break;
		case 'k':
			if(reflections > 1) reflections -= 1;
			break;
		case 'i':
			depth += 0.1f;
			break;
		case 'o':
			depth -= 0.1f;
			break;
		default:
			break;
	}
}


// --------------------------------------------------------
//
// Shader Stuff
//

bool setupShaders()
{
	// Init shader
	initShader.init();
	initShader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/pass.vert");
	initShader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/init.frag");

	initShader.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");

	initShader.prepareProgram();
	printf("InfoLog\n%s\n\n", initShader.getAllInfoLogs().c_str());

	// Shader for ray tracer computations
	shader.init();
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/pass.vert");
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/rayTracer.frag");

	shader.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");
	
	shader.prepareProgram();
	printf("InfoLog\n%s\n\n", shader.getAllInfoLogs().c_str());
	
	// Send our generated data from init shader to our ray tracer
	shader.setUniform("raysOriginsRead", 0);
	shader.setUniform("raysDirectionsRead", 1);
	shader.setUniform("colorTextureRead", 2);

	float lightPos[4] = {2.5f, 4.0f, 5.0f, 1.f};
	shader.setUniform("lightPos", lightPos);

	// Final shader
	finalShader.init();
	finalShader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/pass.vert");
	finalShader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/finalize.frag");

	// Set output of our program
	finalShader.setProgramOutput(0, "outputF");
	finalShader.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");
	
	finalShader.prepareProgram();
	printf("InfoLog\n%s\n\n", finalShader.getAllInfoLogs().c_str());
	
	// Send colors computed in ray tracing shader to final shader
	finalShader.setUniform("colorTexture", 2);

	return(finalShader.isProgramLinked());
}

// ------------------------------------------------------------
//
// OpenGL setup
//

void prepareTexture1D(GLuint& textureID, GLvoid* data, GLsizei size)
{
	// Generate texture name bind it and fill it with data
	glGenTextures(1, &textureID);
	glBindTexture(GL_TEXTURE_1D, textureID); 
	glTexImage1D(GL_TEXTURE_1D, 0, GL_R32F, size, 0, GL_RED, GL_FLOAT, data);

	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_NEAREST); 
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
}

void prepareTexture2D(GLuint& textureID)
{
	// Generate texture name bind it, give it empty image
	glGenTextures(1, &textureID);
	glBindTexture(GL_TEXTURE_2D, textureID);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, 2048, 2048, 0, GL_RGBA, GL_FLOAT, 0);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
}

bool init()
{
	// We will render our scene on "canvas": two triangles over whole screen
	// Create the VAO
	glGenVertexArrays(1, &vao);
	glBindVertexArray(vao);

	// Create buffers for our vertex data
	GLuint buffers[2];
	glGenBuffers(2, buffers);

	// Vertex coordinates buffer
	glBindBuffer(GL_ARRAY_BUFFER, buffers[0]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_STATIC_DRAW);
	glEnableVertexAttribArray(VSShaderLib::VERTEX_COORD_ATTRIB);
	glVertexAttribPointer(VSShaderLib::VERTEX_COORD_ATTRIB, 4, GL_FLOAT, 0, 0, 0);

	// Index buffer
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, buffers[1]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(faceIndex), faceIndex, GL_STATIC_DRAW);

	// Unbind the VAO
	glBindVertexArray(0);

	// Double buffer textures and framebuffers
	for(int i = 0; i < 2; i++)
	{
		// Generate texture name bind it, give it empty image
		prepareTexture2D(raysOriginsTexture[i]);
		prepareTexture2D(raysDirectionsTexture[i]);
		prepareTexture2D(colorTexture[i]);

		// Generate name for our framebuffer and bind it
		glGenFramebuffers(1, &dataFramebuffer[i]);
		glBindFramebuffer(GL_FRAMEBUFFER, dataFramebuffer[i]);

		// Connect our framebuffers with textures
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, raysOriginsTexture[i], 0);
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT1, GL_TEXTURE_2D, raysDirectionsTexture[i], 0);
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT2, GL_TEXTURE_2D, colorTexture[i], 0);
	
		// Check if framebuffer is complete
		if(glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
			return false;
	}

	// Unbind texture
	glBindTexture(GL_TEXTURE_2D, 0);
		
	// Unbind our framebuffer
	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	float objectsData[56] = {	0.f,  1.5f,  2.0f, -1.5f, -1.f, 0.f,  0.f, 0.f,	//sphere 1 (isSphere, radius, center(x,y,z), material number )
								0.f,  1.5f, -2.0f,  0.f,  1.f,  1.f,  0.f, 0.f,	//sphere 2
								1.f,  0.0f,  1.0f,  0.f,  0.f, -5.f,  0.f, 1.f,	//bottom plane (isPlane, normal(x,y,z), vertex on plane(x,y,z), material number)
								1.f,  0.0f, -1.0f,  0.f,  0.f,  5.f,  0.f, 1.f,	//upper plane
								1.f,  1.0f,  0.0f,  0.f, -5.f,  0.f,  0.f, 0.f,	//left plane
								1.f, -1.0f,  0.0f,  0.f,  5.f,  0.f,  0.f, 1.f,	//right plane
								1.f,  0.0f,  0.0f,  1.f,  0.f,  0.f, -5.f, 1.f,	//backwards plane
							};

	// Load objects data to 1D texture and attach it to texture unit 3
	glActiveTexture(GL_TEXTURE3);
	prepareTexture1D(dataID, objectsData, 56);
	shader.setUniform("data", 3);


	float materialsData[28] = { 
							0.2f, 0.2f, 0.2f, 1.f,				//ambient
							0.1f, 0.1f, 0.1f, 1.f,				//specular
							0.90f, 0.90f, 0.90f, 1.f,				//diffuse
							700.f, 0.f,								//shininess, reflection coeficient
							0.15f, 0.15f, 0.15f, 1.f,
							0.6f, 0.6f, 0.6f, 1.f,
							0.4f, 0.4f, 0.4f, 1.f,
							800.f, reflectionCoef,
							};
	
	// Load objects data to 1D texture and attach it to texture unit 4
	glActiveTexture(GL_TEXTURE4);
	prepareTexture1D(materialsID, materialsData, 28);	
	shader.setUniform("materials", 4);

	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);

	return true;
}

void initVSL()
{
//	Init VSML
	vsml = VSMathLib::getInstance();
	vsml->setUniformBlockName("Matrices");
	vsml->setUniformName(VSMathLib::PROJ_VIEW_MODEL, "m_pvm");
}


// ------------------------------------------------------------
//
// Main function
//

int main(int argc, char **argv)
{
//  GLUT initialization
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGBA);

	glutInitContextVersion(3, 3);
	glutInitContextProfile(GLUT_CORE_PROFILE);
	glutInitContextFlags(GLUT_DEBUG);

	glutInitWindowPosition(100,100);
	glutInitWindowSize(width, height);
	glutCreateWindow("PV227 - Whitted Raytracing");


//  Callback Registration
	glutDisplayFunc(renderScene);
	glutReshapeFunc(changeSize);
	glutIdleFunc(renderScene);

//	Mouse and Keyboard Callbacks
	glutKeyboardFunc(processKeys);

//	Return from main loop
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

//	Init GLEW
	glewExperimental = GL_TRUE;
	glewInit();

	printf("Vendor: %s\n", glGetString(GL_VENDOR));
	printf("Renderer: %s\n", glGetString(GL_RENDERER));
	printf("Version: %s\n", glGetString(GL_VERSION));
	printf("GLSL: %s\n", glGetString(GL_SHADING_LANGUAGE_VERSION));
	int param;
	glGetIntegerv(GL_CONTEXT_PROFILE_MASK, &param);
	if(param == GL_CONTEXT_CORE_PROFILE_BIT)
		printf("Context Profile: Core\n");
	else
		printf("Context Profile: Compatibility\n");

	if(!setupShaders())
		exit(1);

	if(!init())
		exit(1);

	initVSL();

//  GLUT main loop
	glutMainLoop();

	return(0);
}
