#version 330

//#define SHADOW

// Spheres and Planes data
uniform sampler1D data;

// Materials data (ambient, specular, diffuse, shininess)
uniform sampler1D materials;

// Ray information for given fragment
uniform sampler2D raysOriginsRead;
uniform sampler2D raysDirectionsRead;
uniform sampler2D colorTextureRead;

// Output to texture
layout(location = 0) out vec4 raysOriginsWrite;
layout(location = 1) out vec4 raysDirectionsWrite;
layout(location = 2) out vec4 colorTextureWrite;	// Information about calculated color

// Position of light
uniform vec4 lightPos;

// Ray structure
struct ray
{
	vec3 origin;
	vec3 direction;
	float length;
};

const int materialSize = 14;
const float eps = 0.001f;

// Phong lighting function
vec4 countPhong(vec3 eyeDir, vec3 normal, float materialNum, vec3 lightDir);

// Intersection of ray and sphere fuction
float intersectionWithSphere(ray r, float radius, vec3 center);

// Intersection of ray and plane function
float intersectionWithPlane(ray r, vec3 planePoint, vec3 planeNormal);

// Function calculating intersection with every object
bool intersection(ray r, out vec3 intersectionPoint, out vec3 normal, out int materialNumber);

//----------------------------------------------------------------
// TODO:
// 1) Inspect functions "intersectionWithSphere", "intersectionWithPlane" and "countPhong".
// 2) In function intersection:
//	- call the correct intersection function.
//	- react to the intersection.
// 3) In function main:
//	- compute the direction to light (from the point of intersection).
//	- NO SHADOW: compute the color using "countPhong".
//	- SHADOW: create a ray towards the light
//	- SHADOW: check if shadow ray intersects something, compute color accordingly
//	- output the reflected ray's origin and direction
//	- output the interpolated color from the previous iterations and this iteration.
// 4) Explain darkening in the center of the image.
//----------------------------------------------------------------

void main()
{
	// Get information about ray from textures and create ray
	ivec2 texCoords = ivec2(gl_FragCoord.xy);
	
	vec3 rayOrigin = texelFetch(raysOriginsRead, texCoords, 0).xyz;
	float reflCoeff = texelFetch(raysOriginsRead, texCoords, 0).w;
	vec3 rayDirection = texelFetch(raysDirectionsRead, texCoords, 0).xyz;
	float rayLength = texelFetch(raysDirectionsRead, texCoords, 0).w;
	
	vec4 previousColor = texelFetch(colorTextureRead, texCoords, 0);
	
	ray r = ray(rayOrigin, rayDirection, rayLength);
	
	// Set default background color
	vec4 col = vec4(vec3(0.f), 1.f);
	
	// Create outputs for intersection function
	vec3 intersectionPoint;
	vec3 normal;
	int materialNumber;

	// Run intersection function
	if(intersection(r, intersectionPoint, normal, materialNumber))
	{
		// mo - material offset in material texture data
		int mo = materialSize * materialNumber;

		// Compute direction to the light
		vec3 lightDir = normalize(lightPos.xyz - intersectionPoint);
#ifndef SHADOW
		col = countPhong(-r.direction, normal, materialNumber, lightDir);
#else
		// Create ray from point of intersection to light, check for intersection
		vec3 shadowDir = lightDir;

		// Move intersection point slightly to the light so that self-intersection does not occure
		vec3 shadowPoint = intersectionPoint + eps * shadowDir;
		
		// Create shadow ray
		ray shadowRay = ray(shadowPoint, shadowDir, length(lightPos.xyz - intersectionPoint));

		// Unused values to put somewhere output from the intersection function
		int shadowMaterial;
		vec3 shadowHitPoint, shadowNormal;

		// Decide whether the point is in light or in shadow
		if(intersection(shadowRay, shadowHitPoint, shadowNormal, shadowMaterial))
		{
			// Set ambient color of used material in that point
			col = vec4(texelFetch(materials, mo, 0).r, texelFetch(materials, mo + 1, 0).r, texelFetch(materials, mo + 2, 0).r, texelFetch(materials, mo + 3, 0).r);
		}
		else
		{
			// Compute light direction and pass it to phong lighting function
			col = countPhong(-r.direction, normal, materialNumber, shadowDir);
		}
#endif

		// Compute secondary ray direction and origin
		vec3 reflectedRayDir = reflect(rayDirection, normal);

		// Compute Shlicks's approximation of the Fresnel term

		// Compute the reflection coefficient between air and the material's index of refraction
		float reflection = texelFetch(materials, mo + 13, 0).r;
		float R0 = (1.f - reflection) / (1.f + reflection);
		R0 *= R0;
		// Compute the approximation
		// cosTheta is the angle between the viewing direction and the halfway vector between the incident light direction and the viewing direction, i.e. the normal
		float cosTheta = dot(-normalize(rayDirection), normalize(normal));
		float rFi = R0 + (1 - R0)*pow(1.f - cosTheta, 5);

		float weightedRefl = clamp(reflCoeff*rFi, 0.f, 0.99f);
		float reflLength = 1e38f;

		// If reflection is zero set reflection and ray length to zero
		if(reflection <= 0.f || reflCoeff*rFi < eps)
		{
			weightedRefl = 0.f;
			reflLength = 0.f;
		}

		// Save the origin of the reflected ray and the fresnel term
		raysOriginsWrite = vec4(intersectionPoint + eps * reflectedRayDir, weightedRefl);
		raysDirectionsWrite = vec4(reflectedRayDir, reflLength);
	}
	else
	{
		raysOriginsWrite = vec4(0.f);
		raysDirectionsWrite = vec4(0.f);
	}
	
	
	// Output color
	colorTextureWrite = clamp(mix(previousColor, col, reflCoeff), 0.f, 1.f);
}
	
bool intersection(ray r, out vec3 intersectionPoint, out vec3 normal, out int materialNumber)
{
	// Terminated rays cannot have intersections
	if(r.length <= 0.f)
		return false;

	// Set closest hit to the ray length
	float closestHit = r.length;

	// Get data size
	int dataSize = textureSize(data, 0);
	
	// Naive intersection test with all objects - cycle through data
	for(int i = 0; i < dataSize; i += 8)
	{
		float object = texelFetch(data, i, 0).r;

		// Test if next set of data create sphere or plane
		if(object == 0.f)
		{
			// Get sphere data from texture
			float radius = texelFetch(data, i + 1, 0).r;
			
			float centerX = texelFetch(data, i + 2, 0).r;
			float centerY = texelFetch(data, i + 3, 0).r;
			float centerZ = texelFetch(data, i + 4, 0).r;

			vec4 center = vec4(centerX, centerY, centerZ, 1.f);
		
			// Test ray-sphere intersection
			float t = intersectionWithSphere(r, radius, center.xyz);

			// If hit distance t is valid range calculate intersection point, normal and get material number
			if(t > 0.f && t < closestHit)
			{
				intersectionPoint = r.origin + t * r.direction;
				normal = normalize(intersectionPoint - center.xyz);
				materialNumber = int(texelFetch(data, i + 5, 0).r);
				
				closestHit = t;
			}
		}
		else if(object == 1.f)
		{
			// Get plane data from texture
			float normalX = texelFetch(data, i + 1, 0).r;
			float normalY = texelFetch(data, i + 2, 0).r;
			float normalZ = texelFetch(data, i + 3, 0).r;
			
			float vertexX = texelFetch(data, i + 4, 0).r;
			float vertexY = texelFetch(data, i + 5, 0).r;
			float vertexZ = texelFetch(data, i + 6, 0).r;

			vec3 planeNormal = vec3(normalX, normalY, normalZ);
			vec3 planeVertex = vec3(vertexX, vertexY, vertexZ);

			// Test ray-plane intersection
			float t = intersectionWithPlane(r, planeVertex, planeNormal);

			// If hit distance t is valid range calculate intersection point, normal and get material number
			if(t > 0.f && t < closestHit)
			{
				intersectionPoint = r.origin + t * r.direction;
				normal = planeNormal;
				materialNumber = int(texelFetch(data, i + 7, 0).r);

				closestHit = t;
			}
		}
	}
	
	if(closestHit < r.length) 
		return true; // Hit closer than ray length found
	else
		return false; // No intersection closer than ray length found
}

float intersectionWithSphere(ray r, float radius, vec3 center)
{
	// Get vector from ray origin to center of sphere
	vec3 oc = r.origin - center;
	
	float b = 2.f * dot(oc, r.direction);
	float c = dot(oc, oc) - radius * radius;

	float d = b * b - 4.f * c;
	if(d < 0.f)
		return -1.f;
	
	// Compute distances of intersections
	float res1 = (-b - sqrt(d)) / 2.f;
	float res2 = (-b + sqrt(d)) / 2.f;

	// Sort solutions
	if(res1 > res2)
	{
		float tmp = res1;
		res1 = res2;
		res2 = tmp;
	}

	// Return the closest valid intersection
	if(res1 > 0.f)
		return res1;
	else if(res2 > 0.f)
		return res2;
	else
		return -1.f;
}

float intersectionWithPlane(ray r, vec3 planePoint, vec3 planeNormal)
{
	// Project ray direction onto plane normal
	float rayPlane = dot(planeNormal, r.direction);

	// If dot product of plane normal and ray direction is zero, they are parallel, return -1.f
	if(rayPlane == 0.f)
		return -1.f;

	// Project vector from ray origin to point on plane onto plane normal
	float pointPlane = dot(planeNormal, planePoint - r.origin);

	// Divide these two projections to get the distance (negative value if no valid intersection)
	return pointPlane / rayPlane;
}

vec4 countPhong(vec3 eyeDir, vec3 normal, float materialNum, vec3 lightDir)
{
	// mo - material offset in material texture data
	int mo = materialSize * int(materialNum);
	
	// Getting material data from materialData texture
	vec4 ambient = vec4(texelFetch(materials, mo, 0).r, texelFetch(materials, mo + 1, 0).r, texelFetch(materials, mo + 2, 0).r, texelFetch(materials, mo + 3, 0).r);
	vec4 specular = vec4(texelFetch(materials, mo + 4, 0).r, texelFetch(materials, mo + 5, 0).r, texelFetch(materials, mo + 6, 0).r, texelFetch(materials, mo + 7, 0).r);
	vec4 diffuse = vec4(texelFetch(materials, mo + 8, 0).r, texelFetch(materials, mo + 9, 0).r, texelFetch(materials, mo + 10, 0).r, texelFetch(materials, mo + 11, 0).r);
	float shininess = texelFetch(materials, mo + 12, 0).r;

	// Compute phong light for point light
	lightDir = normalize(lightDir);
	normal = normalize(normal);
	eyeDir = normalize(eyeDir);

	float intensity = max(dot(normal, lightDir), 0.f);
	vec4 spec = vec4(0.f);

	if(intensity > 0.f)
	{
		vec3 h = normalize(lightDir + eyeDir);
		float intSpec = max(dot(h, normal), 0.f);
		spec = specular * pow(intSpec, shininess);
	}

	return clamp(ambient + diffuse * intensity + spec, 0.f, 1.f);
}
